package com.github.tDBN.dbn;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * 
 * Extends the general configuration to represent a configuration using the 
 * array description of size (markovLag+1) * n, where in each array there are also
 * the proper parents needed from past slices.
 * 
 * @author zlm
 * @author Tiago Leao
 * 
 */
public class LocalConfiguration extends Configuration {

	protected int[] parentIndices;

	/**
	 * If true, considers the child value when matching an observation with the
	 * current configuration. In this case, N_{ijk} is what is being counted.
	 */
	protected boolean considerChild = true;
	
	/**
	 * Allocates the configuration array and sets all parents and the child to
	 * their first value. All input nodes must lie in the range
	 * [0,attributes.size()[.
	 * 
	 * @param attributes
	 *            list of attributes characterizing the nodes
	 * @param parentNodesPast
	 *            list of parent nodes in t
	 * @param parentNodesPresent
	 *            list of parent nodes in t+1
	 * @param childNode
	 *            child node in t+1
	 */
	public LocalConfiguration(List<Attribute> attributes, int markovLag, List<Integer> parentNodesPast,
			List<Integer> parentNodesPresent, int childNode) {
		super(attributes, markovLag);
		this.reset();

		int n = attributes.size();
		int numParentsPast = (parentNodesPast != null) ? parentNodesPast.size() : 0;
		int numParentsPresent = (parentNodesPresent != null) ? parentNodesPresent.size() : 0;
		int numParents = numParentsPast + numParentsPresent;

		parentIndices = new int[numParents];
		int i = 0;

		if (parentNodesPast != null)
			// parentNodesPast ints are already shifted
			for (Integer parentNode : parentNodesPast)
				parentIndices[i++] = parentNode;

		if (parentNodesPresent != null)
			for (Integer parentNode : parentNodesPresent)
				parentIndices[i++] = parentNode + markovLag * n;

		resetParents();

		this.childNode = childNode;
		resetChild();
	}
	
	/**
	 * For easily cloning a LocalConfiguration object.
	 * 
	 */
	public LocalConfiguration(LocalConfiguration original) {
		super(original);
		this.parentIndices = original.parentIndices.clone();
		this.considerChild = original.considerChild;
	}
	
	/**
	 * Allocates Configuration only based on a list of attributes and the configuration array.
	 * 
	 * @param attributes
	 *            list of attributes characterizing the nodes
	 * @param configuration
	 *            array of integers, each integer representing a value of the proper attribute. -1 means the attribute is not in configuration.
	 */
	public LocalConfiguration(List<Attribute> attributes, int[] configuration) {
		super(attributes, configuration);
	}

	public LocalConfiguration(List<Attribute> attributes, int markovLag, List<Integer> parentNodesPast,
			Integer parentNodePresent, int childNode) {
		this(attributes, markovLag, parentNodesPast, (parentNodePresent != null ? Arrays.asList(parentNodePresent)
				: null), childNode);
	}

	public LocalConfiguration(List<Attribute> attributes, int markovLag, List<Integer> parentNodes, int childNode) {
		this(attributes, markovLag, parentNodes, (List<Integer>) null, childNode);
	}

	/**
	 * Sets whether the child value should be considered when matching an
	 * observation with the current configuration.
	 * 
	 * @param state
	 */
	public void setConsiderChild(boolean state) {
		considerChild = state;
	}

	public boolean matches(int[] observation) {

		int n = attributes.size();

		for (int i = 0; i < configuration.length; i++) {
			if (configuration[i] > -1) {
				if (observation[i] != configuration[i]) {
					if (considerChild || i != childNode + n * markovLag) {
						return false;
					}
				}
			}
		}
		return true;
	}
	
	/**
	 * To ignore the static observations in case they are also given.
	 */
	public boolean matches(int[] observationDyn, int[] observationStatic) {
		return matches(observationDyn);
	}

	/**
	 * Updates the configuration of parents' values by incrementing the current
	 * configuration in lexicographical order. If there isn't a new
	 * configuration, resets the parents' values and returns false.
	 * 
	 * @return true if a new parents' configuration is generated.
	 */
	public boolean nextParents() {

		int n = attributes.size();

		for (int i = 0; i < parentIndices.length; i++) {
			if (++configuration[parentIndices[i]] < attributes.get(parentIndices[i] % n).size()) {
				break;
			} else {
				configuration[parentIndices[i]] = 0;
				if (i == parentIndices.length - 1) {
					resetParents();
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Sets all parents to their first value.
	 */
	public void resetParents() {
		for (int i = 0; i < parentIndices.length; i++) {
			configuration[parentIndices[i]] = 0;
		}
	}

	/**
	 * Increments the child value maintaining the current configuration of
	 * parents' values. If there are no more child values, resets to the first
	 * value and returns false.
	 * 
	 * @return true if a new child value is generated.
	 */
	public boolean nextChild() {

		int n = attributes.size();

		if (++configuration[n * markovLag + childNode] < attributes.get(childNode).size()) {
			return true;
		} else {
			resetChild();
			return false;
		}

	}

	/**
	 * Sets the child node to its first value.
	 */
	public void resetChild() {
		int n = attributes.size();
		configuration[n * markovLag + childNode] = 0;
	}

	public int getParentsRange() {
		if (parentIndices.length == 0) {
			return 0;
		}
		int n = attributes.size();
		int result = 1;
		for (int i = 0; i < parentIndices.length; i++) {
			result *= attributes.get(parentIndices[i] % n).size();
		}
		return result;
	}

	public int getChildRange() {
		return attributes.get(childNode).size();
	}

	/**
	 * Calculates the number of parameters required to specify a distribution,
	 * according to the list of parents and the child.
	 * 
	 * @return the number of parameters to specify
	 */
	public int getNumParameters() {
		return getParentsRange() * (getChildRange() - 1);
	}

	public static void main(String[] args) {

		Observations o = new Observations(args[0]);
		List<Integer> parentNodesPast = new ArrayList<Integer>();
		parentNodesPast.add(1);
		parentNodesPast.add(2);
		Integer parentNodePresent = 3;
		int childNode = 1;
		LocalConfiguration c = new LocalConfiguration(o.getAttributes(), 1, parentNodesPast, parentNodePresent,
				childNode);

		System.out.println(o);

		int count = 0;

		do {
			do {
				System.out.println(c);
				count++;
			} while (c.nextChild());
		} while (c.nextParents());

		System.out.println(count);
		System.out.println(c.getNumParameters());
	}

}
